/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Wez Furlong <wez@php.net>                                    |
  +----------------------------------------------------------------------+
*/

/* $Id: epoll.c,v 1.1 2004/12/13 04:56:31 wez Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_event.h"
#include "php_event_int.h"

#if defined(HAVE_EPOLL)

#include <sys/poll.h>
#include <sys/epoll.h>
#include <fcntl.h>

struct epoll_state {
	int epfd;
};

static void *epoll_event_init(TSRMLS_D)
{
	int epfd;
	struct epoll_state *st;
	
	st = ecalloc(1, sizeof(*st));

	epfd = epoll_create(FD_SETSIZE);
	if (epfd == -1) {
		efree(st);
		return NULL;
	}

	fcntl(epfd, F_SETFD, 1); /* close on exec */

	st->epfd = epfd;
	
	return st;
}

static void epoll_event_fini(void *arg TSRMLS_DC)
{
	struct epoll_state *st = (struct epoll_state*)arg;
	close(st->epfd);
	efree(st);
}

static int epoll_event_recalc(void *arg, int max TSRMLS_DC)
{
	struct epoll_state *st = (struct epoll_state*)arg;
	
	return 0;
}

static int epoll_event_add(void *arg, struct php_event *ev TSRMLS_DC)
{
	struct epoll_state *st = (struct epoll_state*)arg;
	int op = EPOLL_CTL_ADD;
	struct epoll_event epev;

	if (ev->engine_data == st) {
		op = EPOLL_CTL_MOD;
	}

	epev.events = 0;
	if (ev->ev_events & EV_READ) {
		epev.events |= POLLIN;
	}
	if (ev->ev_events & EV_WRITE) {
		epev.events |= POLLOUT;
	}
	if (ev->ev_events & EV_EXCEP) {
		epev.events |= POLLPRI;
	}
	epev.data.ptr = ev;

	if (epoll_ctl(st->epfd, op, ev->ev_fd, &epev) == 0) {
		ev->engine_data = st;
		return 0;
	}

	return -1;
}

static int epoll_event_del(void *arg, struct php_event *ev TSRMLS_DC)
{
	struct epoll_state *st = (struct epoll_state*)arg;

	if (ev->engine_data == st) {
		struct epoll_event epev;
		epev.data.ptr = ev;
		if (epoll_ctl(st->epfd, EPOLL_CTL_DEL, ev->ev_fd, &epev) == 0) {
			ev->engine_data = 0;
			return 0;
		}
		return -1;
	}
	return 0;
}

static int epoll_event_dispatch(void *arg, struct timeval *tv TSRMLS_DC)
{
	struct epoll_state *st = (struct epoll_state*)arg;
	int res, i, count, msec, nfds;
	struct php_event *ev;
	struct epoll_event events[32];
	struct epoll_event *pfd;

	msec = tv ? (tv->tv_sec * 1000) + (tv->tv_usec / 1000) : -1;
	count = epoll_wait(st->epfd, events, sizeof(events)/sizeof(events[0]), msec);	

	if (count == -1) {
		if (errno != EINTR) {
			return -1;
		}
		return 0;
	}

	if (count == 0) {
		return 0;
	}

	for (i = 0; i < count; i++) {
		pfd = &events[i];

		res = 0;
		if (pfd->events & (POLLERR|POLLHUP)) {
			res |= EV_READ|EV_WRITE;
		}
		if (pfd->events & POLLIN) {
			res |= EV_READ;
		}
		if (pfd->events & POLLOUT) {
			res |= EV_WRITE;
		}

		ev = pfd->data.ptr;
		res &= ev->ev_events;

		if (res == 0) {
			continue;
		}

		event_active(ev, res, 1 TSRMLS_CC);			
	}
	return 0;	
}

const struct php_eventop php_event_epoll_engine = {
	"epoll",
	epoll_event_init,
	epoll_event_fini,
	epoll_event_add,
	epoll_event_del,
	epoll_event_recalc,
	epoll_event_dispatch
};

#endif

